"""
configuration parameters for the doit file
"""
import os, platform
from ConfigParser import SafeConfigParser as ConfigParser


class Config(object):
    """
    ConfigParser wrapper to handle configuration easily
    """
    VARIABLES = (('pkg_dir', str, 'pkg'),
                 ('distros', list, platform.dist()[2]),
                 ('ppas', list, ''),
                 ('extension', str, 'tmpl'),
                 ('changelog-sed', str, ''),
                 ('tag', bool, 'false'),
                 )

    NAMES = set([name for name, _, _ in VARIABLES])
    TYPES = dict([(name, type)
                  for name, type, default in VARIABLES])
    DEFAULTS = dict([(name, default)
                     for name, type, default in VARIABLES])

    def __init__(self, config_filename='~/.packager.cfg'):
        self.config_filename = os.path.expanduser(config_filename)

        # Provide sensible defaults in case configuration
        # file doesn't have some of them
        self.parser = ConfigParser(self.DEFAULTS)
        self.dirty = False


    def read(self):
        """
        Read configuration file
        """
        if os.path.isfile(self.config_filename):
            self.parser.read(self.config_filename)
        return self


    def get_defaults(self):
        """
        Return default option values
        """
        defaults = {}

        for option in self.parser.defaults().iterkeys():
            if option in self.NAMES:
                defaults[option] = self.get_option('DEFAULT', option)

        return defaults


    def get_option(self, section, option):
        """
        Get the value for an option
        """
        if not self.TYPES.has_key(option):
            raise UnknownOption
        option_type = self.TYPES[option]

        if option_type == str:
            value = self.parser.get(section, option)
        elif option_type == list:
            value = self.parser.get(section, option).split()
        elif option_type == bool:
            value = self.parser.getboolean(section, option)
        else:
            raise UnknownType('Unknown type {type!r} for option {option!r}'
                              .format(option=option,
                                      type=option_type))
        return value


    def remove_option(self, section, option):
        """
        Remove an option from a given section
        """
        self.parser.remove_option(section, option)
        self.dirty = True


    def get_options(self, section):
        """
        Get a dictionary of options for a given section
        """
        if not self.parser.has_section(section):
            return self.get_defaults()

        options = dict([(option, self.get_option(section, option))
                        for option in self.parser.options(section)
                        if option in self.NAMES])

        return options


    def set_option(self, section, option, new_value):
        """
        Set specific value for option
        """
        # Create section if it doesn't exist yet
        if section != 'DEFAULT' and not self.parser.has_section(section):
            self.parser.add_section(section)
            self.dirty = True

        if not self.TYPES.has_key(option):
            raise UnknownOption
        option_type = self.TYPES[option]

        if option_type == str :
            # No transformation needed for this type
            pass
        elif option_type == bool:
            new_value = str(new_value)
        elif option_type == list:
            if not isinstance(new_value, str):
                new_value = ' '.join(new_value)
        else:
            raise UnknownType('Unknown type {type!r} for option {option!r}'
                              .format(option=option,
                                      type=option_type))

        if self.parser.has_option(section, option):
            value = self.parser.get(section, option)
            if value != new_value:
                self.parser.set(section, option, new_value)
                self.dirty=True
        else:
            self.parser.set(section, option, new_value)
            self.dirty=True


    def set_options(self, section, options):
        """
        Set a dictionary of options for a given sections
        Mark instance as dirty if changes are made to values
        """
        for option, new_value in options.items():
            self.set_option(section, option, new_value)

        return self


    def write(self):
        """
        Write changes to configuration file
        """
        if self.dirty:
            with open(self.config_filename, 'w') as f:
                self.parser.write(f)
                self.dirty = False


class UnknownOption(Exception):
    pass

class UnknownType(Exception):
    pass
