"""
doit file to generate source and binary packages automatically
"""
import sys, os, subprocess, shutil, platform

from mako.template import Template
import apt.debfile, debian.changelog

from packager.info import DebianInfo, FileInfo, BzrInfo
from packager.config import Config

DEBIAN_INFO = DebianInfo()
BZR_INFO = BzrInfo()
CONFIG = Config().read().get_options(DEBIAN_INFO.source)
FILE_INFO = FileInfo(DEBIAN_INFO, CONFIG)


def render_template(template_file, destination_file, extra_context=None):
    """
    Render a template file passing to the template common values
    as available replacement patterns
    """
    template = Template(filename=template_file)
    context = FILE_INFO.get_template_context()
    if extra_context:
        context.update(extra_context)
    output = template.render(**context)

    dirname = os.path.dirname(destination_file)
    if dirname and not os.path.isdir(dirname):
        os.makedirs(dirname)
    with open(destination_file, 'w') as f:
        f.write(output)

    return True


def task_code():
    """
    Create package directory and copy source files
    ('bzr branch' not used to take into account uncommited changes)
    """
    def copy_file(from_file, to_file):
        """
        Copy file to packaging directory
        keeping all their attributes
        """
        dirname = os.path.dirname(to_file)
        if dirname and not os.path.isdir(dirname):
            os.makedirs(dirname)
        shutil.copyfile(from_file, to_file)
        shutil.copystat(from_file, to_file)
        return True


    def update_changelog(filename, distro):
        """
        Update changelog entries
        to generate the interim version to be uploaded
        """
        comment_line = ('# Automatically modified by packager tool '
                        '(https://launchpad.net/packager)\n')

        # Avoid updating changelog file
        # when it has already been updated
        with open(filename) as f:
            line = f.readline()
        if line == comment_line:
            return

        with open(filename) as f:
            changelog = debian.changelog.Changelog(f)

        # Replace distribution
        changelog.distributions = distro

        # Replace version (if configured)
        if CONFIG['changelog-sed']:
            context = FILE_INFO.get_basic_template_context()
            context['distro'] = distro
            new_version = FILE_INFO.get_version(context)
            changelog.version = debian.changelog.Version(new_version)

        with open(filename, 'w') as f:
            f.write(comment_line)
            changelog.write_to_open_file(f)

        return True

    dependencies=BZR_INFO.versioned_files()
    targets=FILE_INFO.destination_files

    for target, dependency in zip(targets, dependencies):
        is_template = FILE_INFO.is_template(dependency)
        action = copy_file if not is_template else render_template

        # Changelog is modified after being copied
        # to generate the right version of the package
        # to be installed/uploaded
        if target == FILE_INFO.destination_changelog:
            target_distro = os.environ.get('PACKAGER_TARGET_DISTRO',
                                           platform.dist()[2])

            yield {'name': dependency,
                   'actions': [(action, (dependency, target)),
                               (update_changelog,
                                (target, target_distro))],
                   'file_dep': [dependency],
                   'targets': [target]}
        else:
            yield {'name': dependency,
                   'actions': [(action, (dependency, target))],
                   'file_dep': [dependency],
                   'targets': [target]}

    yield {'name': 'clean',
           'actions': None,
           'clean': ['rm -rf {dir}'.format(dir=CONFIG['pkg_dir'])]}


def task_render():
    """
    Render templates using values internally available
    """
    dependencies = FILE_INFO.template_files
    targets = FILE_INFO.source_rendered_template_files

    for target, dependency in zip(targets, dependencies):
        yield {'name': dependency,
               'actions': [(render_template, (dependency, target))],
               'file_dep': [dependency],
               'targets': [target],
               'clean': True}


def task_source():
    """
    Create source package from source distribution
    """
    def create_source_package():
                          subprocess.check_call('debuild --no-tgz-check -S',
                                  cwd=FILE_INFO.package_dir,
                                  shell=True)

    return {'actions': [(create_source_package,)],
            'file_dep': FILE_INFO.destination_files,
            'targets': FILE_INFO.source_package_files,
            'clean': True}


def upload(distro):
    """
    Upload package for the given distro to the configured PPAs
    """
    environment = os.environ.copy()
    environment['PACKAGER_TARGET_DISTRO'] = distro
    subprocess.check_call('doit --dir {directory} -f {dodo_file} raw_upload'
                          .format(directory=os.getcwd(),
                                  dodo_file=__file__),
                          shell=True,
                          env=environment)


def task_raw_upload():
    """
    Upload source package to multiple PPAs (internal use task)
    """
    def raw_upload(ppa):
        subprocess.check_call('dput -f {ppa} {filename}'
                              .format(ppa=ppa,
                                      filename=FILE_INFO.upload_file),
                              cwd=CONFIG['pkg_dir'],
                              shell=True)

    ppas = CONFIG['ppas']
    for ppa in ppas:
        yield {'name': ppa,
               'actions': [(raw_upload, (ppa,))],
               'file_dep': FILE_INFO.source_package_files}


def task_check_changes():
    """
    Check if there are uncommited changes in the branch
    """
    def check_changes():
        if BZR_INFO.uncommitted_changes():
            sys.stderr.write('Please commit all your changes '
                             'before uploading a new package version\n')
            return False

        return True

    return {'actions': [check_changes],
            'verbosity': 2}


def task_upload():
    """
    Upload source package to multiple PPAs and distros
    """
    yield {'name': 'check_changes',
           'actions': None,
           'task_dep': ['check_changes']}

    distros = CONFIG['distros']
    for distro in distros:
        yield {'name': distro,
               'actions': [(upload, (distro,))]}

    if CONFIG['tag']:
        yield {'name': 'tag',
               'actions': ['bzr tag --force {version}'
                           .format(version=DEBIAN_INFO.version)]}


def task_binary():
    """
    Create binary packages from the source distribution
    """
    def create_binary_package():
        subprocess.check_call('debuild -b -uc -us',
                              cwd=FILE_INFO.package_dir,
                              shell=True)

    return {'actions': [(create_binary_package,)],
            'file_dep': FILE_INFO.destination_files,
            'targets': FILE_INFO.binary_package_files,
            'clean': True}


def need_to_install():
    """
    Check if it's needed to install binary packages

    Binary packages will installed when:
    - the .deb file for any of them is missing
    - the binary package is currently uninstalled
    - the installed version isn't the same one as in the .deb file
    """
    for binary_package_file in FILE_INFO.binary_package_files:
        if not os.path.isfile(binary_package_file):
            return True

        binary_package = apt.debfile.DebPackage(binary_package_file)
        result = binary_package.compare_to_version_in_cache()
        if result != binary_package.VERSION_SAME:
            return True

    return False


def task_install():
    """
    Install binary package locally
    """
    def install():
        command = 'gksudo' if 'DISPLAY' in os.environ else 'sudo'
        subprocess.check_call('{command} debi'.format(command=command),
                              cwd=FILE_INFO.package_dir,
                              shell=True)

    return {'actions': [(install,)],
            'file_dep': FILE_INFO.binary_package_files,
            'uptodate': [not need_to_install()]}
