"""
Retrieve information used to calculate dependencies and targets in the doit file
"""
import os, re, subprocess, platform
from datetime import datetime
from ConfigParser import SafeConfigParser as ConfigParser

from mako.template import Template

import debian.changelog
import apt_pkg

from .interface import PackagerException


class DebianInfo(object):
    """
    Information parsed from debian files
    (debian/changelog, debian/control)
    """
    CONTROL='debian/control'
    CHANGELOG='debian/changelog'
    CHANGELOG_MAINTAINER=re.compile(r'^(?P<name>(?:\s*\w+)+)\s+(?P<email><.*>)')

    def __init__(self, directory=None):
        if directory is None:
            self.directory = os.getcwd()
            self.control = self.CONTROL
            self.changelog = self.CHANGELOG
        else:
            self.directory = directory
            self.control = os.path.join(self.directory, self.CONTROL)
            self.changelog = os.path.join(self.directory, self.CHANGELOG)

        self.parse()


    def parse(self):
        """
        Parse debian files

        Parsing is already done on object instantiation,
        this method should be called only when files
        have been updated
        """
        self._parse_control()
        self._parse_changelog()


    def _parse_changelog(self):
        """
        Parse debian/changelog using the source package name
        from  debian/control file
        """
        if not os.path.isfile(self.changelog):
            raise PackagerException('Changelog file not found:\n{0}'
                                    .format(os.path.relpath(self.control,
                                                            self.directory)))


        parser = debian.changelog.Changelog(open(self.changelog))

        try:
            self.version = str(parser.version)
            self.distro = parser.distributions
            maintainer = parser.author
        except:
            raise PackagerException('Error while parsing changelog file:\n'
                                    '{0}'.format(os.path.relpath(self.changelog,
                                                                 self.directory)))

        self.source_package_name = ('{source}-{version}'
                                    .format(source=self.source,
                                            version=self.version))

        match = self.CHANGELOG_MAINTAINER.match(maintainer)
        self.maintainer_name = match.group('name')
        self.maintainer_email = match.group('email')


    def _parse_control(self):
        """
        Parse control file
        """
        if not os.path.isfile(self.control):
            raise PackagerException('Control file not found:\n{0}'
                                    .format(os.path.relpath(self.control,
                                                            self.directory)))


        def check_section_keys(section, keys):
            for key in keys:
                if key not in section:
                    raise PackagerException('{0} line not found in control file:\n'
                                            '{1}'
                                            .format(key, os.path.relpath(self.control,
                                                                         self.directory)),
                                            'Failed section:\n{0}'.format(section))

        with open(self.control) as f:
            sections = [section for section in apt_pkg.TagFile(f)]

        if not sections:
            raise PackagerException('No sections found in control file:\n{0}'
                                    .format(os.path.relpath(self.control,
                                                            self.directory)))

        # Get source package data
        source_section = sections[0]
        check_section_keys(source_section, ('Source', 'Build-Depends'))

        self.source = source_section['Source']
        self.build_depends = source_section['Build-Depends']

        # Get binary packages data
        binary_sections = sections[1:]
        for section in binary_sections:
            check_section_keys(section, ('Package', 'Architecture'))

        self.packages = [section['Package']
                         for section in binary_sections]
        self.architectures = [section['Architecture']
                              for section in binary_sections]


class FileInfo(object):
    """
    Information extracted regarding the filesystem
    """
    MACHINE_TO_PKG = {'x86_64': 'amd64',
                      'i686': 'i386',
                      'armv7l': 'armel'}


    def __init__(self, debian_info, cfg):
        # Get values from configuration
        # and version control
        self.debian_info=debian_info
        self.bzr_info=BzrInfo(debian_info.directory)
        self.pkg_dir = cfg['pkg_dir']
        self.extension = cfg['extension']
        self.changelog_sed = cfg['changelog-sed']

        # Generate values based on the previously read configuration
        target_distro = os.environ.get('PACKAGER_TARGET_DISTRO',
                                       platform.dist()[2])

        # Binary version string that will be generated
        # for the target distro
        context = self.get_basic_template_context()
        context['distro'] = target_distro
        self.version = self.get_version(context)

        # Source package name that will be generated
        # for the current distro
        self.source_package_name = self.get_source_package_name(context)

        # Directory where the source files
        # for the package should be copied to
        self.package_dir = os.path.join(self.pkg_dir, self.source_package_name)

        # Destination files that should be used
        # to create the source and binary packages
        def get_destination_filename(filename):
            root = os.path.splitext(filename)[0]
            return filename if not self.is_template(filename) else root

        versioned_files = self.bzr_info.versioned_files()
        self.destination_files = [os.path.join(self.pkg_dir,
                                               self.source_package_name,
                                               get_destination_filename(versioned_file))
                                  for versioned_file in versioned_files]

        # Changelog file location in destination files
        self.destination_changelog = os.path.join(self.pkg_dir,
                                                  self.source_package_name,
                                                  self.debian_info.changelog)

        # Files that must exist after source package has been created
        source_package_filename_prefix = ('{directory}/{source}_{version}'
                                          .format(directory=self.pkg_dir,
                                                  source=self.debian_info.source,
                                                  version=self.version))
        self.source_package_files = ('{prefix}.dsc\n'
                                     '{prefix}_source.changes\n'
                                     '{prefix}_source.build\n'
                                     '{prefix}.tar.gz\n'
                                     .format(prefix=source_package_filename_prefix)).split()

        # File that must be passed to dput to upload a file
        self.upload_file = '%s_%s_source.changes' % (self.debian_info.source, self.version)

        # Files that must exist after binary packages have been created
        self.binary_package_files = self._get_binary_package_files()

        # Template files under version control
        self.template_files = [filename for filename in versioned_files
                               if self.is_template(filename)]

        # Template files locally rendered
        self.source_rendered_template_files = [os.path.splitext(template_file)[0]
                                               for template_file in self.template_files]

        # Template files rendered in the packaging directory
        self.package_rendered_template_files = [os.path.join(self.pkg_dir,
                                                             self.debian_info.source_package_name,
                                                             os.path.splitext(template_file)[0])
                                                for template_file in self.template_files]


    def is_template(self, filename):
        """
        Return True if filename is a template file
        based on its extension
        """
        extension = os.path.splitext(filename)[1]
        return (extension == '.{0}'.format(self.extension))


    def get_basic_template_context(self):
        """
        Return context to be used just in changelog version
        replacement template
        """
        return {'revno': self.bzr_info.revno(),
                'date': datetime.now(),
                'distro': platform.dist()[2]}


    def get_template_context(self):
        """
        Return context to be used in templates
        """
        context = self.get_basic_template_context()
        context['version'] = self.version
        return context


    def get_version(self, context):
        """
        Return the binary version that will be generated
        for a given distro
        """
        # Apply user provided replacement to package version
        if self.changelog_sed:
            process = subprocess.Popen(['sed', '-e', self.changelog_sed],
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE)
            stdout, _ = process.communicate(input=self.debian_info.version)

            # Use sed output only if no problem was found
            # (that is, expression is not malformed)
            if process.returncode == 0:
                version_template = stdout
            else:
                version_template = self.debian_info.version
            version = Template(version_template).render(**context)
        else:
            version = self.debian_info.version

        return version


    def get_source_package_name(self, context):
        """
        Return the source package name that will be generated
        for a given distro
        """
        package_name = ('{source}-{version}'
                        .format(source=self.debian_info.source,
                                version=self.get_version(context)))
        return package_name


    def _get_binary_package_files(self):
        """
        Return files that must exist after binary packages has been created
        """
        build_architecture = self.MACHINE_TO_PKG[platform.machine()]

        files = []
        packages = self.debian_info.packages
        architectures = self.debian_info.architectures

        for package, architecture in zip(packages, architectures):
            if architecture == 'all':
                file_architecture = 'all'
            elif architecture == 'any':
                file_architecture = build_architecture
            elif architecture != build_architecture:
                continue

            binary_package_file = ('{directory}/{package}_{version}_{architecture}.deb'
                                   .format(directory=self.pkg_dir,
                                           package=package,
                                           version=self.version,
                                           architecture=file_architecture))
            files.append(binary_package_file)

        return files


class BzrInfo(object):
    def __init__(self, directory=None):
        if directory is None:
            directory = os.getcwd()

        self.directory = directory


    def is_branch(self):
        """
        Return True if directory is a branch
        """
        return bool(self.revno())


    def missing_changes(self):
        """
        Return True if there are missing changes
        Return False if parent branch is unknown
        or branches are synchronized
        """
        process = subprocess.Popen(['bzr', 'missing'],
                                   stdout=subprocess.PIPE,
                                   cwd=self.directory)
        process.communicate()
        return process.returncode == 1


    def revno(self):
        """
        Return branch revision number
        """
        revno=(subprocess.Popen(['bzr', 'revno'],
                               stdout=subprocess.PIPE,
                               cwd=self.directory)
               .communicate()[0].rstrip())
        return revno


    def versioned_files(self):
        """
        Return files under bazaar version control
        (this is includes files with uncommited changes
        because it's useful to test changes locally
        without committing them)
        """
        paths=(subprocess.Popen(['bzr', 'ls', '-R', '-V'],
                               stdout=subprocess.PIPE,
                               cwd=self.directory)
               .communicate()[0].splitlines())

        # Filter out directories
        files=[path for path in paths if not os.path.isdir(path)]
        return files


    def uncommitted_changes(self):
        """
        Return true if there are uncommitted changes
        """
        output = (subprocess.Popen(['bzr', 'status', '-SV'],
                                  stdout = subprocess.PIPE,
                                  cwd=self.directory)
                  .communicate()[0])
        return bool(output)


class GPGInfo(object):
    EMAIL=re.compile('<.*>')

    def emails(self):
        emails = []
        process = subprocess.Popen(['gpg', '--list-secret-keys'],
                                   stdout=subprocess.PIPE)
        stdout, _ = process.communicate()
        for line in stdout.splitlines():
            match = self.EMAIL.search(line)
            if match:
                emails.append(match.group(0))

        return emails


class DputInfo(ConfigParser):
    CONFIG_FILENAME='~/.dput.cf'
    FQDN = 'ppa.launchpad.net'
    METHOD = 'ftp'
    LOGIN = 'anonymous'
    ALLOW_UNSIGNED_UPLOADS = '0'

    # Names must be at least two characters long and start with a letter or number.
    # All letters must be lower-case. The characters +, - and . are also allowed after the first character.
    VALID_NAME = r'[a-z0-9][a-z0-9\+\.\-]+'
    INCOMING = re.compile(r'^~{name}/{name}/ubuntu/?$'.format(name=VALID_NAME))

    def __init__(self):
        ConfigParser.__init__(self)
        self.filename = os.path.expanduser(self.CONFIG_FILENAME)
        if os.path.isfile(self.filename):
            self.read(self.filename)

    def is_ppa(self, ppa_name):
        if not self.has_section(ppa_name):
            return False

        options = self.options(ppa_name)
        for option in ['fqdn', 'method', 'login', 'allow_unsigned_uploads']:
            if not (option in options
                    and self.get(ppa_name, option) == getattr(self, option.upper())):
                return False

        if not ('incoming' in options
                and self.INCOMING.match(self.get(ppa_name, 'incoming'))):
            return False

        return True

