"""
CLI interface
"""
import sys, os
from . import parser
from ... import config, info, interface

class Packager(interface.PackagerInterface):
    def main(self, args):
        """
        Check arguments and launch doit properly
        """
        for filename in self.DEBIAN_FILES:
            self.check_file(filename)

        # Read changelog file
        self.debian_info = info.DebianInfo()
        self.cfg = config.Config().read()

        source = self.debian_info.source
        options = self.cfg.get_options(source)
        self.file_info = info.FileInfo(self.debian_info, options)
        self.parser = parser.Parser(self.BUILD, self.DESCRIPTION, options)

        args = self.parser.parse_args(args)
        self.update_config_file(args)

        method_name = ('handle_{subcommand}_cmd'
                       .format(subcommand=args.subcommand))
        if hasattr(self, method_name):
            getattr(self, method_name)(args)
        else:
            self.handle_doit_cmd(args)
        return 0


    def check_file(self, filename):
        """
        Make sure that a file exists
        """
        if not os.path.isfile(filename):
            self.parser = parser.Parser(self.BUILD, self.DESCRIPTION)
            self.parser.print_usage(sys.stderr)
            self.parser._print_message(('{prog}: error: {filename!r} file not found\n'
                                        .format(prog=self.parser.prog, filename=filename)),
                                       sys.stderr)
            sys.exit(1)


    def handle_config_cmd(self, args):
        """
        Handle config subcommand
        """
        method_name = ('handle_config_{subcommand}_cmd'
                       .format(subcommand=args.config_subcommand))
        getattr(self, method_name)(args)


    def handle_config_show_cmd(self, args):
        """
        Handle config show subcommand
        """
        source = self.debian_info.source
        options = self.cfg.get_options(source)

        if args.default:
            print ('defaults\n'
                   '--------')
            display_options = self.cfg.get_defaults()
        else:
            version = self.debian_info.version
            header = ('{source} - ({version})'
                      .format(source=source, version=version))
            print header
            print '-'*len(header)
            defaults = self.cfg.get_defaults()

            display_options = {}
            for key, value in options.iteritems():
                if key in defaults and defaults[key] == value:
                    display_key = 'd ' + key
                else:
                    display_key = '  ' + key

                display_options[display_key] = value

        display_options = sorted(display_options.iteritems(),
                                 key=lambda opt: opt[0])
        display_options_len = max([len(key) for key, _ in display_options])
        for key, value in display_options:
            print ('{key:{key_len}}: {value!r}'
                   .format(key=key, key_len=display_options_len, value=value))


    def handle_config_set_cmd(self, args):
        """
        Handle config set subcommand
        """
        if args.default:
            section = 'DEFAULT'
        else:
            section = self.debian_info.source

        try:
            self.cfg.set_option(section, args.name, args.value)
        except config.UnknownOption:
            sys.stderr.write('Unknown option {option!r}.\n'
                             'Please use one of the following options:\n{options}\n'
                             .format(option=args.name,
                                     options='\n'.join(['- {0}'.format(name)
                                                        for name in sorted(self.cfg.NAMES)])))
        else:
            self.cfg.write()


    def handle_config_remove_cmd(self, args):
        """
        Handle config remove subcommand
        """
        source = self.debian_info.source

        if args.default:
            self.cfg.remove_option('DEFAULT', args.name)
        else:
            self.cfg.remove_option(source, args.name)
        self.cfg.write()


    def handle_info_cmd(self, args):
        """
        Handle help subcommand
        """
        source = self.debian_info.source
        binary_packages = self.debian_info.packages
        changelog_version = self.debian_info.version
        changelog_distro = self.debian_info.distro

        options = self.cfg.get_options(source)
        distros = options['distros']
        ppas = options['ppas']

        context = self.file_info.get_basic_template_context()
        versions = []
        for distro in distros:
            context['distro'] = distro
            versions.append(self.file_info.get_version(context))

        display_options = [('source package', source),
                           ('binary packages', binary_packages),
                           ('changelog version', changelog_version),
                           ('changelog distro', changelog_distro),
                           ('target PPAs', ppas),
                           ('target distributions', distros),
                           ('target versions', versions)]

        display_options_len = max([len(key) for key, _ in display_options])

        for key, value in display_options:
            print ('{key:{key_len}}: {value!r}'
                   .format(key=key, key_len=display_options_len, value=value))


    def handle_help_cmd(self, args):
        """
        Handle help subcommand
        """
        # Print subcommand help only if passed and known
        if args.command and args.command in args.subparsers:
            args.subparsers[args.command].print_help()
        # Print general help otherwise
        else:
            self.parser.print_help()


    def handle_doit_cmd(self, args):
        """
        Run command through doit
        """
        # Launch doit passing the selected subcommand
        # as a task to the dodo file
        from packager import dodo
        from doit.doit_cmd import cmd_main as doit_main
        doit_main(['--dir', os.getcwd(),
                   '-f', dodo.__file__,
                   args.subcommand])


    def update_config_file(self, args):
        """
        Take care of the configuration values passed through the command line
        """
        if args.subcommand == 'upload':
            self.update_config_file_upload(args)

        self.cfg.write()


    def update_config_file_upload(self, args):
        """
        Check that all required arguments for upload subcommand
        are available either through the command line
        or through the configuration file
        """
        source = self.debian_info.source
        options = self.cfg.get_options(source)

        for key in ('tag', 'ppas', 'distros'):
            value = getattr(args, key)
            if value is not None:
                options[key] = value
            elif key not in options:
                args.subparsers[args.subcommand].print_help()

                sys.stderr.write('\nERROR: No {key!r} passed through the command line\n'
                                 'and none found in the config file {filename!r}.\n'
                                 'Please make sure that a value can be loaded '
                                 'from any of those sources.\n'
                                 .format(key=key, filename=self.cfg.config_filename))
                sys.exit(1)

        # Update configuration file with values passed
        # through the command line
        self.cfg.set_options(source, options)

