"""
Packager GTK interface module
"""
import pygtk
pygtk.require("2.0")
import gtk, gobject, glib, gio

# Use threads
gtk.gdk.threads_init()

import sys, os, logging, tempfile

import argparse

from ... import info, interface
from . import window, install, upload, about, util

# "Steal" icon from synaptic
gtk.window_set_default_icon_name('synaptic')


class Packager(window.GladeWindow, interface.PackagerInterface):
    DEFAULT_WINDOW_TITLE = 'Packager'
    CLOSE_WIDGETS = ('close_menuitem', 'close_toolbutton')
    ACTION_WIDGETS = ('install_menuitem', 'install_toolbutton',
                      'upload_menuitem', 'upload_toolbutton')

    # Register new signals
    gobject.signal_new('opened', gtk.Window, gobject.SIGNAL_RUN_FIRST, None, (str,))
    gobject.signal_new('closed', gtk.Window, gobject.SIGNAL_RUN_FIRST, None, ())

    def __init__(self):
        """
        Create user interface widgets
        """
        # Parse command line arguments
        self.args = self._parse_cmd_line()

        # Load widgets
        super(Packager, self).__init__('packager.glade')
        self._set_accel_group()

        # Replace default exception handler
        sys.excepthook = self._exception_handler

        self._configure_logging(self.args.log_level, self.args.log_file)
        logging.info(self.BUILD)
        logging.info(self.args)

        # Initialize internal data
        self.directory=None
        self.debian_info=None
        self.bzr_info=None

        self.file_monitors = []
        self.update_scheduled = False

        if self.args.directory:
            glib.idle_add(self._open_directory, self.args.directory)

        self.window.show_all()


    def _set_accel_group(self):
        """
        Set accelerators for common actions
        """
        group = gtk.AccelGroup()
        self.window.add_accel_group(group)

        accelerators_data = [
            (self.open_menuitem, '<Control>o'),
            (self.close_menuitem, '<Control>w'),
            (self.quit_menuitem, '<Control>q'),
            (self.install_menuitem, '<Control>i'),
            (self.upload_menuitem, '<Control>u'),
            ]

        for menuitem, keys in accelerators_data:
            accel_key, accel_mod = gtk.accelerator_parse(keys)
            menuitem.add_accelerator('activate',
                                     group,
                                     accel_key,
                                     accel_mod,
                                     gtk.ACCEL_VISIBLE)


    def _parse_cmd_line(self):
        """
        Parse command line arguments
        """
        log_levels = ['debug', 'info', 'warning', 'error', 'critical']
        parser = argparse.ArgumentParser(description=self.DESCRIPTION)
        parser.add_argument('-d', '--directory', type=os.path.realpath,
                            help='Branch directory that should be opened')
        parser.add_argument('-l', '--log-level',
                            dest='log_level',
                            choices=log_levels,
                            default='warning',
                            help=('Log level. One of {0} or {1} '
                                  '(%(default)s by default)'
                                  .format(', '.join(log_levels[:-1]), log_levels[-1])))
        parser.add_argument('-f', '--log-file', type=os.path.realpath,
                            dest='log_file',
                            help='Filename where logs will be written')
        parser.add_argument('-v', '--version', action='version',
                            version=self.BUILD,
                            help='Display version and exit')

        args = parser.parse_args()
        args.log_level = getattr(logging, args.log_level.upper())
        if args.log_file is None:
            _, args.log_file = tempfile.mkstemp(prefix='packager-', suffix='.log')


        return args


    def _configure_logging(self, log_level, log_filename):
        """
        Configure logging based on the configuration
        passed through the command line
        """
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)

        # Log to sys.stderr using log level
        # passed through command line
        log_handler = logging.StreamHandler()
        formatter = logging.Formatter('%(levelname)s: %(message)s')
        log_handler.setFormatter(formatter)
        log_handler.setLevel(log_level)
        logger.addHandler(log_handler)

        # Log to file using DEBUG log level
        log_directory = os.path.dirname(log_filename)
        if not os.path.exists(log_directory):
            os.makedirs(log_directory)

        log_handler = logging.FileHandler(log_filename)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s: '
                                      '%(message)s')
        log_handler.setFormatter(formatter)
        log_handler.setLevel(logging.DEBUG)
        logger.addHandler(log_handler)


    def _exception_handler(self, type, value, traceback):
        """
        Log and display all exceptions
        """
        def handle_response(dialog, response):
            """
            Report bug through apport
            """
            dialog.destroy()

            if response == util.ExceptionDialog.REPORT_BUG:
                self.report_bug()

        def display_dialog():
            """
            Display custom dialog with exception information
            """
            dialog = util.ExceptionDialog(exc_data=(type, value, traceback))
            dialog.connect('response', handle_response)
            dialog.show_all()

        # Display dialog later in the main thread
        glib.idle_add(display_dialog)


    def report_bug(self):
        """
        Report bug using apport
        """
        module = {}
        APPORT_GTK = '/usr/share/apport/apport-gtk'
        execfile(APPORT_GTK, module)
        sys.argv = [APPORT_GTK]
        ui = module['GTKUserInterface']()
        ui.options.package = 'packager'
        ui.options.directory = self.directory
        ui.options.log_filename = self.args.log_file

        ui.run_report_bug()

        # Set again default icon
        # (apport interface changes it)
        gtk.window_set_default_icon_name('synaptic')


    def open_cb(self, widget):
        """
        Open project root directory
        """
        dialog = gtk.FileChooserDialog('Open project root directory',
                                       buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                                  gtk.STOCK_OPEN, gtk.RESPONSE_ACCEPT),
                                       action = gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)

        response = dialog.run()
        directory = dialog.get_filename()
        dialog.destroy()

        if response != gtk.RESPONSE_ACCEPT:
            return False

        self._open_directory(directory)
        return False


    def _open_directory(self, directory):
        """
        Make sure that directory contains all packaging file needed
        and open it
        """
        logging.debug('Opening {0!r}...'.format(directory))

        # Check that it's a bazaar branch
        bzr_info = info.BzrInfo(directory)
        if not bzr_info.is_branch():
            message = 'Directory is <b>not</b> a bazaar branch:\n{0}'.format(directory)
            util.MessageDialog('Error', message).run()
            return

        # Check that required files are in the directory
        missing_filenames = []
        for filename in self.DEBIAN_FILES:
            full_filename = os.path.join(directory, filename)
            if not os.path.isfile(full_filename):
                missing_filenames.append(filename)

        if missing_filenames:
            message = ('The following files have <b>not</b> been found:\n'
                       '{0}'.format('\n'.join(missing_filenames)))
            util.MessageDialog('Error', message).run()
            return

        # Set internal state
        self.directory=directory
        self.debian_info=info.DebianInfo(directory=directory)
        self.bzr_info=bzr_info

        # Monitor debian file to update information accordingly
        for filename in (self.debian_info.changelog,
                         self.debian_info.control):
            monitor = gio.File(filename).monitor()
            monitor.connect('changed', self.file_changed_cb)
            self.file_monitors.append(monitor)

        self.window.emit('opened', directory)


    def opened_cb(self, window, directory):
        """
        Fill widgets data
        """
        logging.info('Directory opened: {0!r}'.format(directory))

        # Update window title
        self.window.set_title('{0} - {1}'.format(directory,
                                                 self.DEFAULT_WINDOW_TITLE))

        # Enable/Disable widgets properly
        for widget in self.CLOSE_WIDGETS + self.ACTION_WIDGETS:
            getattr(self, widget).set_sensitive(True)

        self._update_debian_info()


    def _update_debian_info(self):
        """
        Fill widgets with data extracted from debian files
        """
        self.source.set_text(self.debian_info.source)
        self.version.set_text(self.debian_info.version)

        self.binary.clear()
        for binary_package in self.debian_info.packages:
            self.binary.append((binary_package,))


    def file_changed_cb(self, monitor, gfile, other_file, event_type):
        """
        Update interface when debian files are updated
        """
        if self.update_scheduled:
            return

        glib.timeout_add_seconds(1, self.update_timeout_cb, gfile)
        self.update_scheduled = True


    def update_timeout_cb(self, gfile):
        """
        Update information from debian files
        """
        logging.debug('Updating information from debian files...')
        try:
            self.debian_info.parse()
        except interface.PackagerException:
            message = ('Parsing error while updating information from file:\n'
                       '{0}\n\nOpened directory will be <b>closed</b>.'
                       .format(os.path.relpath(gfile.get_path(), self.directory)))

            def handle_response(dialog, response):
                dialog.destroy()
                if response == util.ExceptionDialog.REPORT_BUG:
                    self.report_bug()
                self.window.emit('closed')

            dialog = util.ExceptionDialog(message)
            dialog.connect('response', handle_response)
            dialog.show_all()

        self.update_scheduled = False
        self._update_debian_info()


    def close_cb(self, widget):
        """
        Close project root directory
        """
        self.window.emit('closed')


    def closed_cb(self, window):
        """
        Clean widgets data
        """
        logging.info('Directory closed: {0!r}'.format(self.directory))

        # Stop monitoring debian files
        for monitor in self.file_monitors:
            monitor.cancel()
        self.file_monitors = []

        # Update window title
        self.window.set_title(self.DEFAULT_WINDOW_TITLE)

        # Enable/Disable widgets properly
        for widget in self.CLOSE_WIDGETS + self.ACTION_WIDGETS:
            getattr(self, widget).set_sensitive(False)

        # Reset widgets to default values
        self.source.set_text('')
        self.version.set_text('')
        self.binary.clear()

        # Reset other internal data
        self.directory=None
        self.debian_info=None
        self.bzr_info=None


    def quit_menuitem_activate_cb(self, menuitem):
        """
        Exit from application
        """
        logging.debug('Quit')
        self.window.destroy()


    def window_destroy_cb(self, window):
        """
        Quit from application
        """
        logging.debug('Destroy')
        gtk.main_quit()


    def install_cb(self, widget):
        """
        Install binary packages
        """
        install.InstallAssistant(self.window, self.debian_info, self.DODO_FILE)


    def upload_cb(self, widget):
        """
        Upload source package to launchpad
        """
        upload.UploadAssistant(self.window, self.debian_info, self.DODO_FILE)


    def report_problem_cb(self, menuitem):
        """
        Report problem through apport
        """
        self.report_bug()


    def about_cb(self, window):
        """
        Display about dialog
        """
        dialog = about.AboutDialog()
        dialog.run()


    def main(self, args):
        """
        Run GTK main loop
        """
        gtk.main()
