import pygtk
pygtk.require("2.0")
import gtk, glib

import logging, subprocess

from . import window, util, apt_util

class Assistant(window.GladeWindow):
    # Some variables common to all the assistants implemented
    CHECKBUTTONS = ('replace_ubuntu_checkbutton',
                    'append_revno_checkbutton')


    SED_TO_CHECKBUTTONS =  \
        { 's/ubuntu/${distro}/g': (True, False),
          's/ubuntu/${distro}/g;s/$/~rev${revno}/': (True, True),
          's/$/~rev${revno}/': (False, True),
          '': (False, False),
          }

    CHECKBUTTONS_TO_SED = dict([(value, key)
                                for key, value in SED_TO_CHECKBUTTONS.items()])

    PULSE_TIMEOUT = 250


    def __init__(self, parent, ui_file, dodo_file):
        self.dodo_file = dodo_file
        self.view_log_button = None

        # Working thread to run commands in the background
        # while a progress bar is being pulsed
        self.thread = None

        # Load and display widgets
        super(Assistant, self).__init__(ui_file)
        self.assistant.set_transient_for(parent)
        self.assistant.show_all()


    def _update_target_version_helper(self):
        """
        Get target version based on checkbuttons status
        """
        options = tuple([getattr(self, checkbutton).get_active()
                     for checkbutton in self.CHECKBUTTONS])
        changelog_sed = self.CHECKBUTTONS_TO_SED[options]
        self.project_config['changelog-sed'] = changelog_sed

        process = subprocess.Popen(['sed', '-e', changelog_sed],
                                   stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE)
        stdout, _ = process.communicate(input=self.debian_info.version)

        # Use sed output only if no problem was found
        # (that is, expression is not malformed)
        if process.returncode == 0:
            version_template = stdout
        else:
            version_template = self.debian_info.version

        return version_template


    def cancel_cb(self, assistant):
        """
        Destroy assistant on cancel
        """
        logging.debug('Assistant {0!r} cancelled'
                      . format(assistant.get_title()))
        self.assistant.destroy()


    def close_cb(self, assistant):
        """
        Destroy assistant on close
        """
        logging.debug('Assistant {0!r} closed'
                      . format(assistant.get_title()))
        self.assistant.destroy()


    def prepare_cb(self, assistant, page):
        """
        Fill widgets with data properly
        before page is shown to the user
        """
        # Verify that extra button is removed if it exists
        if self.view_log_button:
            assistant.remove_action_widget(self.view_log_button)
            self.view_log_button.destroy()
            self.view_log_button = None

        # Get page number and deliver data to page handler
        page_number=assistant.get_current_page()
        page_handler_name = self.PAGE_HANDLERS[page_number]
        page_handler = getattr(self, page_handler_name)

        # TBD: Investigate strange logging behaviour
        # with multiple threads
        logger = logging.getLogger('packager')
        logger.debug('Assistant: {0!r}, Page: {1!r}, {2}, {3}'
                     .format(assistant.get_title(),
                             assistant.get_page_title(page),
                             page_number,
                             page_handler))
        self.page_number = page_number
        page_handler(assistant, page)


    def prepare_build_dependencies_error_page(self, assistant, label):
        """
        Display error message and abort installation
        """
        satisfiable, package = self.check_thread.target_return
        if not satisfiable:
            build_depends = [dependency.strip()
                             for dependency in self.debian_info.build_depends.split(',')]

            message = ("It's <b>not</b> possible to satisfy build dependencies "
                       'for your package:\n{build_depends}\n\n'
                       'To fix this problem, please make sure that:\n'
                       '- Package names are <b>not</b> misspelled\n'
                       '- <b>All</b> archives needed to get build dependencies are available'
                       .format(build_depends='\n'.join(build_depends)))
        else:
            unauthenticated = package.required_changes[2]

            if not unauthenticated:
                raise ValueError(unauthenticated)

            unauthenticated_str = '\n'.join(['- {0}'.format(package_name)
                                             for package_name in unauthenticated])
            message = ("It's <b>not</b> possible to authenticate "
                       'the following packages:\n{0}\n\n'
                       'To fix this, please review your <b>sources</b> configuration'
                       .format(unauthenticated_str))

        label.set_markup(message)


    def prepare_build_dependencies_confirmation_page(self, assistant, label):
        """
        Prompt user to install build dependences
        """
        package = self.check_thread.target_return[1]
        install, remove = package.required_changes[:2]

        message = ['The following changes will be made to meet the source package '
                   'build dependencies:\n']
        for section, changes in (('Install', install), ('Remove', remove)):
            if changes:
                changes_str = '\n'.join(['- {0}'.format(change)
                                         for change in changes])
                message.append('<b>{0}</b>:\n{1}'.format(section, changes_str))

        label.set_markup('\n'.join(message))
        assistant.set_page_complete(label, True)


    def prepare_build_dependencies_progress_page(self, assistant, page):
        """
        Check that source package build dependencies are met
        """
        package = self.check_thread.target_return[1]
        install, remove = package.required_changes[:2]

        # Create command to be executed in a separate thread
        command_data = ['gksudo', '-D', 'Packager', '--',
                        'python', apt_util.__file__]
        if install:
            install_option = ['-i']
            install_option.extend(install)
            command_data.extend(install_option)
        if remove:
            remove_option = ['-d']
            remove_option.extend(remove)
            command_data.extend(remove_option)

        command = util.Command(command_data)
        progressbar = self.build_dependencies_progressbar

        def apply_changes_complete():
            """
            Wait for build dependencies to be installed
            before setting page complete so that user can move forward
            """
            if command.process.returncode == 0:
                next_page = self.forward_page(assistant.get_current_page())
                assistant.set_current_page(next_page)
                return

            assistant.set_page_type(page, gtk.ASSISTANT_PAGE_SUMMARY)
            self.view_log_button.set_sensitive(True)
            message = 'Changes application failure'
            progressbar.set_text(message)

        progressbar.set_text('Applying changes...')
        self._run_command(command.run,
                          progressbar, apply_changes_complete)


    def _prepare_version_helper(self):
        """
        Helper for the different prepare version methods
        in child classes
        """
        self.changelog_version.set_text(self.debian_info.version)

        # Set checkbuttons status based on configured changelog patterns
        changelog_sed=self.project_config['changelog-sed']
        options = self.SED_TO_CHECKBUTTONS.get(changelog_sed, (False, False))

        for checkbutton_name, option in zip(self.CHECKBUTTONS, options):
            checkbutton = getattr(self, checkbutton_name)
            with util.block_handler(checkbutton, self.version_checkbutton_toggled_cb):
                checkbutton.set_active(option)


    def _run_command(self, command, progressbar, callback):
        """
        Run a command in a separate thread while progressbar is being pulsed
        and execute a callback when the command execution finishes
        """
        # Set page to default values
        assistant = self.assistant
        page = assistant.get_nth_page(assistant.get_current_page())
        assistant.set_page_complete(page, False)
        assistant.set_page_type(page, gtk.ASSISTANT_PAGE_PROGRESS)

        # Add view log button if it doesn't exist
        if self.view_log_button is None:
            self.view_log_button = gtk.Button('View log...')
            self.view_log_button.set_sensitive(False)
            assistant.add_action_widget(self.view_log_button)
            self.view_log_button.show()
            self.view_log_button.connect('clicked', self._display_command_results)

        self.thread = util.Thread(target=command)
        self.thread.start()
        glib.timeout_add(self.PULSE_TIMEOUT,
                         self.pulse_target, self.thread, progressbar,
                         callback)


    def _display_command_results(self, button):
        """
        Display doit output in a separate dialog
        """
        command = self.thread.target_return
        returncode = command.process.returncode

        # Create custom dialog
        # to display possibly long results in a scrolled window
        dialog = gtk.Dialog('View log',
                            buttons = (gtk.STOCK_OK, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.set_property('border-width', 10)
        dialog.vbox.set_spacing(5)
        hbox = gtk.HBox(spacing=5)
        image = gtk.image_new_from_stock((gtk.STOCK_DIALOG_INFO
                                          if returncode == 0
                                          else gtk.STOCK_DIALOG_ERROR),
                                         gtk.ICON_SIZE_DIALOG)
        image.set_alignment(0.5, 0.0)
        label = gtk.Label()
        label.set_markup('Command execution <b>{0}</b>'
                         .format('success'
                                 if returncode == 0
                                 else 'failure'))
        hbox.pack_start(image, expand=False)
        hbox.pack_start(label, expand=False)
        dialog.vbox.pack_start(hbox, expand=False)

        table_data = [('Return code', str(returncode))]
        if command.stdout:
            table_data.append(('Standard output', command.stdout))
        if command.stderr:
            table_data.append(('Standard error', command.stderr))

        table = gtk.Table(rows=len(table_data),
                          columns=2, homogeneous=False)
        table.set_row_spacings(5)
        table.set_col_spacings(5)
        for row_number, (label_str, data_str) in enumerate(table_data):
            if '\n' in data_str:
                expander = gtk.Expander(label_str)
                scrolled_window = gtk.ScrolledWindow()
                scrolled_window.set_policy(gtk.POLICY_AUTOMATIC,
                                           gtk.POLICY_AUTOMATIC)
                textview = gtk.TextView()
                textview.set_editable(False)
                buffer = textview.get_buffer()
                buffer.set_text(data_str)
                scrolled_window.add(textview)
                expander.add(scrolled_window)
                table.attach(expander, 0, 2, row_number, row_number+1)
            else:
                label = gtk.Label('{0}:'.format(label_str))
                label.set_alignment(1.0, 0.5)
                table.attach(label, 0, 1, row_number, row_number+1, gtk.FILL, gtk.FILL)

                if len(data_str) > 40:
                    entry = gtk.Entry()
                    entry.set_text(data_str)
                    entry.set_property('editable', False)
                    table.attach(entry, 1, 2, row_number, row_number+1, yoptions=gtk.FILL)
                else:
                    label = gtk.Label(data_str)
                    label.set_alignment(0.0, 0.5)
                    label.set_selectable(True)
                    table.attach(label, 1, 2, row_number, row_number+1, yoptions=gtk.FILL)

        dialog.vbox.pack_start(table)
        dialog.connect('response', lambda w, d: dialog.destroy())

        # Make ok button grab focus
        # so that enter key defaults to clicking in the button
        ok_button = dialog.get_action_area().get_children()[0]
        ok_button.grab_focus()

        dialog.show_all()
        dialog.run()


    def pulse_target(self, thread, progressbar,
                     callback, callback_args=()):
        """
        Pulse progress while doit process is running
        """
        if thread.is_alive():
            progressbar.pulse()
            return True

        glib.idle_add(callback, *callback_args)
        return False
