import pygtk
pygtk.require("2.0")
import gtk

import platform, logging
from datetime import datetime

import apt.debfile
from mako.template import Template

from . import assistant, util, apt_util
from ... import info, config

class InstallAssistant(assistant.Assistant):
    """
    Assistant wrapper for the local installation of binary package
    """
    PAGE_INTRODUCTION = 0
    PAGE_BUILD_DEPENDENCIES_ERROR = 1
    PAGE_BUILD_DEPENDENCIES_CONFIRMATION = 2
    PAGE_BUILD_DEPENDENCIES_PROGRESS = 3
    PAGE_VERSION_SELECTION = 4
    PAGE_BINARY_PROGRESS = 5
    PAGE_INSTALL_PROGRESS = 6
    PAGE_DEPENDENCIES_ERROR = 7
    PAGE_DEPENDENCIES_CONFIRMATION = 8
    PAGE_DEPENDENCIES_PROGRESS = 9

    # Create a list of page handlers that will be called
    # to prepare the page they are in charge of
    PAGE_HANDLERS = ('prepare_introduction_page',
                     'prepare_build_dependencies_error_page',
                     'prepare_build_dependencies_confirmation_page',
                     'prepare_build_dependencies_progress_page',
                     'prepare_version_selection_page',
                     'prepare_binary_progress_page',
                     'prepare_install_progress_page',
                     'prepare_dependencies_error_page',
                     'prepare_dependencies_confirmation_page',
                     'prepare_dependencies_progress_page')

    def __init__(self, parent, debian_info, dodo_file):
        self.debian_info = debian_info
        self.bzr_info = info.BzrInfo(debian_info.directory)
        self.config = config.Config().read()
        self.project_config = self.config.get_options(self.debian_info.source)
        self.file_info = info.FileInfo(debian_info, self.project_config)

        # Load and display widgets
        super(InstallAssistant, self).__init__(parent, 'install.glade', dodo_file)
        self.assistant.set_forward_page_func(self.forward_page, None)

    def forward_page(self, current_page, data=None):
        """
        Control flow of assistant by selecting next page to display
        """
        page = self.assistant.get_nth_page(current_page)
        logging.debug('Forward page: {0} ({1})'
                      .format(self.assistant.get_page_title(page),
                              current_page))

        # Check that build depends are/can be met
        if current_page == self.PAGE_INTRODUCTION:
            self.check_thread.join()
            satisfiable, package = self.check_thread.target_return
            if not satisfiable:
                return self.PAGE_BUILD_DEPENDENCIES_ERROR

            install, remove, unauthenticated = package.required_changes
            if unauthenticated:
                return self.PAGE_BUILD_DEPENDENCIES_ERROR

            if install or remove:
                return self.PAGE_BUILD_DEPENDENCIES_CONFIRMATION

            # When everything is fine,
            # user can choose changelog version parameters
            return self.PAGE_VERSION_SELECTION

        # By default pages are ordered
        # so return just next one
        return current_page+1


    def assistant_destroy_cb(self, assistant):
        """
        Make sure that packages with unmet dependencies
        are removed when closing the dialog
        """
        if self.page_number == self.PAGE_DEPENDENCIES_CONFIRMATION:
            self._remove_binary_packages()

    def _remove_binary_packages(self):
        """
        Remove binary packages when they have unmet dependencies
        in a different thread so that user doesn't need to wait for it
        """
        command_data = ['gksudo', '-D', 'Packager', '--',
                                'python', apt_util.__file__, '-d']

        package_files = [apt.debfile.DebPackage(filename)
                         for filename in self.file_info.binary_package_files]
        package_names = [package_file.pkgname
                         for package_file in package_files]
        for package_name in package_names:
            command_data.append(package_name)
        command = util.Command(command_data)
        thread = util.Thread(target=command.run)
        thread.start()


    def prepare_introduction_page(self, assistant, label):
        """
        Set introduction message according to source package name
        """
        label.set_markup('This assistant will guide you through '
                         'all the configuration details needed '
                         'to generate the binary packages '
                         'from the <b>{0}</b> source code and install them.'
                         .format(self.debian_info.source))

        def check_build_depends_target():
            """
            Check if build dependencies are/can be met
            """
            package = apt.debfile.DscSrcPackage()
            satisfiable = package.satisfy_depends_str(self.debian_info.build_depends)
            return satisfiable, package

        # Check build depends as early as possible
        # so user doesn't have to wait later
        self.check_thread = util.Thread(target=check_build_depends_target)
        self.check_thread.start()

        assistant.set_page_complete(label, True)


    def prepare_version_selection_page(self, assistant, page):
        """
        Set target version according to checkboxes status
        """
        self._prepare_version_helper()
        self.update_target_version()
        assistant.set_page_complete(page, True)


    def prepare_binary_progress_page(self, assistant, page):
        """
        Set the installation message correctly
        and handle the progress bar
        """
        # Create command to be executed in a separate thread
        command_data = ('doit',
                        '--dir', self.debian_info.directory,
                        '-f', self.dodo_file,
                        'binary')
        command = util.Command(command_data)
        progressbar = self.binary_progressbar

        self.uninstallable = []
        self.install = []
        self.remove = []
        self.unauthenticated = []

        def binary_target():
            command.run()

            if command.process.returncode == 0:
                # Check binary files dependencies
                package_files = [apt.debfile.DebPackage(filename)
                                 for filename in self.file_info.binary_package_files]
                package_names = [package_file.pkgname
                                 for package_file in package_files]

                # Check if there's any package
                # that cannot be installed
                self.uninstallable = [package_file for package_file in package_files
                                      if not package_file.check()]

                # Check what kind of changes are required
                # to install the packages properly
                for package_file in package_files:
                    self.install.extend([package_name
                                         for package_name in package_file.required_changes[0]
                                         if package_name not in package_names])
                    self.remove.extend([package_name
                                        for package_name in package_file.required_changes[1]
                                        if package_name not in package_names])
                    self.unauthenticated.extend(package_file.required_changes[2])


                if self.uninstallable or self.unauthenticated:
                    # If there's any uninstallable/unauthenticated package
                    # remove all installed packages
                    self._remove_binary_packages()

                # Remove duplicates
                self.install = set(self.install)
                self.remove = set(self.remove)
                self.unauthenticated = set(self.unauthenticated)

            return command


        def binary_complete():
            """
            Wait for binary packages to be generated
            before setting page complete so that user can move forward
            """
            if command.process.returncode == 0:
                # Display error if dependencies cannot be met
                if self.uninstallable or self.unauthenticated:
                    assistant.set_current_page(self.PAGE_DEPENDENCIES_ERROR)
                    return

                # Meet dependencies for .deb files before installation
                if self.install or self.remove:
                    assistant.set_current_page(self.PAGE_DEPENDENCIES_CONFIRMATION)
                    return

                # Install packages directly
                # if no need to deal with dependences
                assistant.set_current_page(self.PAGE_INSTALL_PROGRESS)
                return

            message = 'Binary packages generation failure'
            assistant.set_page_type(page, gtk.ASSISTANT_PAGE_SUMMARY)
            self.view_log_button.set_sensitive(True)
            progressbar.set_text(message)
            progressbar.set_fraction(1.0)

        # Update configuration file
        self.config.set_options(self.debian_info.source,
                                self.project_config)
        self.config.write()

        logging.debug('Project configuration: {0}'
                      .format(self.project_config))

        progressbar.set_text('Generating {0!r} binary packages...'
                             .format(self.debian_info.source))
        self._run_command(binary_target, progressbar, binary_complete)


    def prepare_install_progress_page(self, assistant, page):
        """
        Install binary packages and take care of their dependencies
        """
        # Create command to be executed in a separate thread
        command_data = ('doit',
                        '--dir', self.debian_info.directory,
                        '-f', self.dodo_file,
                        'install')
        command = util.Command(command_data)
        progressbar = self.install_progressbar

        def install_complete():
            """
            Verify that all packages were correctly installed
            """
            if command.process.returncode == 0:
                message = 'Binary packages installation success'
            else:
                message = 'Binary packages installation failure'

            assistant.set_page_type(page, gtk.ASSISTANT_PAGE_SUMMARY)
            self.view_log_button.set_sensitive(True)
            progressbar.set_text(message)
            progressbar.set_fraction(1.0)

        progressbar.set_text('Installing {0!r} binary packages...'
                             .format(self.debian_info.source))
        self._run_command(command.run, progressbar, install_complete)


    def prepare_dependencies_error_page(self, assistant, label):
        """
        Display known error regarging binary packages dependencies
        """
        if self.uninstallable:
            uninstallable_str = '\n'.join(['- <b>{0}</b>'.format(package_file.pkgname)
                                           for package_file in self.uninstallable])
            message = ("It's <b>not</b> possible to install "
                       'the following packages:\n{0}\n\n'
                       'To fix this, please review the <b>dependencies</b> '
                       'for these packages and your <b>sources</b> configuration'
                       .format(uninstallable_str))
        elif self.unauthenticated:
            unauthenticated_str = '\n'.join(['- <b>{0}</b>'.format(package)
                                             for package in self.unauthenticated])
            message = ("It's <b>not</b> possible to authenticate "
                       'the following packages:\n{0}\n\n'
                       'To fix this, please review your <b>sources</b> configuration'
                       .format(unauthenticated_str))

        label.set_markup(message)


    def prepare_dependencies_confirmation_page(self, assistant, label):
        """
        Confirm that binary packages dependencies should be installed
        """
        message = ['The following changes will be made to meet the binary packages '
                   'dependencies:\n']
        for section, changes in (('Install', self.install), ('Remove', self.remove)):
            if changes:
                changes_str = '\n'.join(['- {0}'.format(change)
                                         for change in changes])
                message.append('<b>{0}</b>:\n{1}'.format(section, changes_str))

        label.set_markup('\n'.join(message))
        assistant.set_page_complete(label, True)


    def prepare_dependencies_progress_page(self, assistant, page):
        """
        Install binary packages dependencies
        """
        # Create command to be executed in a separate thread
        command_data = ['gksudo', '-D', 'Packager', '--',
                        'python', apt_util.__file__]
        if self.install:
            install_option = ['-i']
            install_option.extend(self.install)
            command_data.extend(install_option)
        if self.remove:
            remove_option = ['-d']
            remove_option.extend(self.remove)
            command_data.extend(remove_option)

        command = util.Command(command_data)
        progressbar = self.dependencies_progressbar

        def apply_changes():
            """
            Install changes and remove packages in case of failure
            """
            # Install dependencies
            command.run()

            # Remove installed packages in case of a problem
            if command.process.returncode != 0:
                self._remove_binary_packages()

            return command

        def apply_changes_complete():
            """
            Wait for dependencies to be installed
            before setting page complete so that user can move forward
            """
            if command.process.returncode == 0:
                assistant.set_current_page(self.PAGE_INSTALL_PROGRESS)
                return


            message = 'Changes application failure'
            assistant.set_page_type(page, gtk.ASSISTANT_PAGE_SUMMARY)
            self.view_log_button.set_sensitive(True)
            progressbar.set_text(message)
            progressbar.set_fraction(1.0)

        progressbar.set_text('Applying changes...')
        self._run_command(apply_changes,
                          progressbar, apply_changes_complete)


    def update_target_version(self):
        """
        Update target version based on checkbuttons status
        """
        version_template = self._update_target_version_helper()
        context =  {'revno': self.bzr_info.revno(),
                    'date': datetime.now(),
                    'distro': platform.dist()[2]}
        target_version = Template(version_template).render(**context)
        self.target_version.set_text(target_version)
        logging.debug('Target version: {0}'.format(target_version))


    def version_checkbutton_toggled_cb(self, checkbutton):
        self.update_target_version()
