import pygtk
pygtk.require("2.0")
import gtk, glib

import re, logging
from datetime import datetime

import apt.debfile
from mako.template import Template

from . import assistant, util
from ... import info, config

class UploadAssistant(assistant.Assistant):
    """
    Assistant wrapper for the upload to launchpad of source packages
    for different distributions and versions
    """
    PAGE_INTRODUCTION = 0
    PAGE_CHECK_PROGRESS = 1
    PAGE_CHECK_PROBLEM = 2
    PAGE_BUILD_DEPENDENCIES_ERROR = 3
    PAGE_BUILD_DEPENDENCIES_CONFIRMATION = 4
    PAGE_BUILD_DEPENDENCIES_PROGRESS = 5
    PAGE_DISTRIBUTIONS = 6
    PAGE_VERSIONS = 7
    PAGE_PPAS = 8
    PAGE_SOURCE_PROGRESS = 9
    PAGE_UPLOAD_PROGRESS = 10


    # Create a list of page handlers that will be called
    # to prepare the page they are in charge of
    PAGE_HANDLERS = ('prepare_introduction_page',
                     'prepare_check_progress_page',
                     'prepare_check_problem_page',
                     'prepare_build_dependencies_error_page',
                     'prepare_build_dependencies_confirmation_page',
                     'prepare_build_dependencies_progress_page',
                     'prepare_distributions_page',
                     'prepare_versions_page',
                     'prepare_ppas_page',
                     'prepare_source_progress_page',
                     'prepare_upload_progress_page')

    DISTRO = 0
    SELECTED = 1

    PPA = 0
    COLOR = 1

    # Names must be at least two characters long and start with a letter or number.
    # All letters must be lower-case. The characters +, - and . are also allowed after the first character.
    VALID_NAME = r'[a-z0-9][a-z0-9\+\.\-]+'
    PPA_PATTERN = re.compile(r'ppa:{name}/{name}'.format(name=VALID_NAME))

    # Row used to let the user know that treeview can be edited
    DEFAULT_PPA_ROW = ('ppa:<owner>/<name>', 'gray')

    # Keys used to delete rows in models
    DELETE_KEYS = [gtk.accelerator_parse(key_name)[0]
                   for key_name in ('Delete', 'BackSpace')]

    def __init__(self, parent, debian_info, dodo_file):
        self.debian_info = debian_info
        self.bzr_info = info.BzrInfo(debian_info.directory)
        self.dput_info = info.DputInfo()
        self.config = config.Config().read()
        self.project_config = self.config.get_options(self.debian_info.source)

        # Load and display widgets
        super(UploadAssistant, self).__init__(parent, 'upload.glade',
                                              dodo_file)
        self.assistant.set_forward_page_func(self.forward_page, None)


    def forward_page(self, current_page, data=None):
        """
        Control flow of assistant by selecting next page to display
        """
        page = self.assistant.get_nth_page(current_page)
        logging.debug('Forward page: {0} ({1})'
                      .format(self.assistant.get_page_title(page),
                              current_page))

        if (current_page == self.PAGE_INTRODUCTION
            or current_page == self.PAGE_CHECK_PROGRESS):
            if self.check_thread.is_alive():
                return self.PAGE_CHECK_PROGRESS

            error, package = self.check_thread.target_return
            if isinstance(error, str):
                return self.PAGE_CHECK_PROBLEM

            satisfiable = error
            if not satisfiable:
                return self.PAGE_BUILD_DEPENDENCIES_ERROR

            install, remove, unauthenticated = package.required_changes
            if unauthenticated:
                return self.PAGE_BUILD_DEPENDENCIES_ERROR

            if install or remove:
                return self.PAGE_BUILD_DEPENDENCIES_CONFIRMATION

            # When everything is fine,
            # user can select the target distributions
            return self.PAGE_DISTRIBUTIONS

        return current_page+1


    def prepare_introduction_page(self, assistant, label):
        """
        Customize introduction page depending on source package name
        """
        label.set_markup('This assistant will guide you through '
                         'all the configuration details needed '
                         'to generate the source package '
                         'from the <b>{0}</b> source code '
                         'and upload it for diferent distributions '
                         'and PPAs in launchpad'
                         .format(self.debian_info.source))

        def check_target():
            if self.bzr_info.uncommitted_changes():
                return 'uncommitted-changes', None

            if self.bzr_info.missing_changes():
                return 'missing-changes', None

            if self.debian_info.maintainer_email not in info.GPGInfo().emails():
                return 'maintainer-email', None

            package = apt.debfile.DscSrcPackage()
            satisfiable = package.satisfy_depends_str(self.debian_info.build_depends)
            return satisfiable, package

        self.check_thread = util.Thread(target=check_target)
        self.check_thread.start()

        assistant.set_page_complete(label, True)


    def prepare_check_progress_page(self, assistant, page):
        """
        Wait for check thread to finish
        """
        def check_complete():
            """
            Set the next page immediately
            without using clicking on forward button
            """
            next_page = self.forward_page(assistant.get_current_page())
            assistant.set_current_page(next_page)

        glib.timeout_add(self.PULSE_TIMEOUT,
                         self.pulse_target, self.check_thread, self.check_progressbar,
                         check_complete)


    def prepare_check_problem_page(self, assistant, label):
        """
        Display error based on detected problem
        """
        error, package = self.check_thread.target_return

        if error == 'uncommitted-changes':
            assistant.set_page_title(label, 'Uncommitted changes')
            message = ('Bazaar branch has <b>uncommitted</b> changes.\n\n'
                       'Please <b>commit</b> and <b>push</b> all your changes '
                       'before uploading a new package version.')
        elif error == 'missing-changes':
            assistant.set_page_title(label, 'Missing changes')
            message = ('Bazaar branch is <b>not</b> synchronized '
                       'with its parent branch.\n\n'
                       'Please <b>pull</b>/<b>push</b> all missing changes '
                       'before uploading a new package version.')
        elif error == 'maintainer-email':
            assistant.set_page_title(label, 'Maintainer identity mismatch')
            escape = glib.markup_escape_text
            message = ('According to <b>changelog</b>, last modification was made by:\n'
                       '{name} {email}\n\n'
                       'This is <b>not</b> reported as any of your gpg identities, '
                       'so you will be unable to sign the package before upload.\n\n'
                       'To fix the problem:\n'
                       '- Add a description of your changes using <b>dch</b>/<b>debchange</b>\n'
                       '- <b>Commit</b> the change\n'
                       '- Try to upload the source package again'
                       .format(name=escape(self.debian_info.maintainer_name),
                               email=escape(self.debian_info.maintainer_email)))
        else:
            raise ValueError(error)

        label.set_markup(message)


    def prepare_distributions_page(self, assistant, page):
        """
        Populate selected distros according to project config
        """
        for row in self.distros:
            distro = row[self.DISTRO]
            if distro.lower() in self.project_config['distros']:
                row[self.SELECTED] = True
                assistant.set_page_complete(page, True)

        logging.debug('Target distributions: {0}'
                      .format(' '.join([distro
                                        for distro, selected in self.distros
                                        if selected])))

    def prepare_versions_page(self, assistant, page):
        """
        Set target versions according to checkboxes status
        """
        # Save selected distros from previous page
        # in project config
        distros=[distro.lower()
                 for distro, selected in self.distros if selected]
        self.project_config['distros']=distros

        self._prepare_version_helper()
        self._update_target_versions()


    def prepare_ppas_page(self, assistant, page):
        """
        Help the user set the target PPAs properly
        """
        ppas = self.project_config['ppas']
        self.ppas.clear()
        for ppa in ppas:
            color = 'black' if self._validate_ppa_name(ppa) else 'red'
            self.ppas.append((ppa, color))
        self.ppas.append(self.DEFAULT_PPA_ROW)
        self._update_ppas_page_complete()


    def prepare_source_progress_page(self, assistant, page):
        """
        Generate source package
        """
        # Create command to be executed in a separate thread
        command_data = ('doit',
                        '--dir', self.debian_info.directory,
                        '-f', self.dodo_file,
                        'source')
        command = util.Command(command_data)
        progressbar = self.source_progressbar

        def source_complete():
            """
            Wait for the source package to be completed
            and move to next page
            """
            if command.process.returncode == 0:
                assistant.set_current_page(self.PAGE_UPLOAD_PROGRESS)
                return

            message = 'Source package generation failure'
            assistant.set_page_type(page, gtk.ASSISTANT_PAGE_SUMMARY)
            self.view_log_button.set_sensitive(True)
            progressbar.set_text(message)
            progressbar.set_fraction(1.0)

        # Update configuration file
        self.config.set_options(self.debian_info.source,
                                self.project_config)
        self.config.write()

        logging.debug('Project configuration: {0}'
                      .format(self.project_config))
        progressbar.set_text('Generating {0!r} source package...'
                             .format(self.debian_info.source))
        self._run_command(command.run, progressbar, source_complete)


    def prepare_upload_progress_page(self, assistant, page):
        """
        Upload source package to launchpad
        using selected configuration parameters
        """
        # Create command to be executed in a separate thread
        command_data = ('doit',
                        '--dir', self.debian_info.directory,
                        '-f', self.dodo_file,
                        'upload')
        command = util.Command(command_data)
        progressbar = self.upload_progressbar

        def upload_complete():
            """
            Wait for the source package to be upload
            and display result
            """
            if command.process.returncode == 0:
                message = 'Source package upload success'
            else:
                message = 'Source package upload failure'
            assistant.set_page_type(page, gtk.ASSISTANT_PAGE_SUMMARY)
            self.view_log_button.set_sensitive(True)
            progressbar.set_text(message)
            progressbar.set_fraction(1.0)

        progressbar.set_text('Uploading {0!r} source package...'
                             .format(self.debian_info.source))
        self._run_command(command.run, progressbar, upload_complete)


    def ppa_keypressed_cb(self, widget, event):
        """
        Remove row is any delete key was pressed
        """
        # Make sure that key pressed
        # is one that means delete
        if event.keyval not in self.DELETE_KEYS:
            return

        selection = self.ppas_treeview.get_selection()
        model, iterator = selection.get_selected()

        # Make sure that a row is selected
        if not iterator:
            return

        # Make sure that selected row
        # isn't the last one
        # (the one used as helper to add new rows)
        row_number = model.get_path(iterator)[0]
        if row_number == len(model)-1:
            return

        model.remove(iterator)
        self._update_ppas_page_complete()


    def ppa_editing_started_cb(self, renderer, editable, path):
        """
        Set page complete status to false  until editing is finished
        """
        self.assistant.set_page_complete(self.ppas_vbox, False)


    def ppa_editing_canceled_cb(self, renderer):
        """
        Set page complete status to the one before editing started
        """
        self._update_ppas_page_complete()


    def ppa_edited_cb(self, renderer, path, new_ppa_name):
        """
        Validate entered PPA and update model
        """
        path = int(path)
        self.ppas[path] = (new_ppa_name,
                           'black' if self._validate_ppa_name(new_ppa_name) else 'red')


        if (path == len(self.ppas)-1
            and new_ppa_name != self.DEFAULT_PPA_ROW[self.PPA]):
            self.ppas.append(self.DEFAULT_PPA_ROW)

        self._update_ppas_page_complete()


    def _validate_ppa_name(self, ppa_name):
        """
        Validate PPA name either by looking at .dput.cf
        or by checking that name matches pattern
        """
        if self.dput_info.is_ppa(ppa_name):
            return True

        if self.PPA_PATTERN.match(ppa_name):
            return True

        return False


    def _update_ppas_page_complete(self):
        """
        Set PPAs page complete based on ppa name validation results
        """
        # Update project config with current ppas name value
        ppas = [ppa_name for ppa_name, color in self.ppas][:-1]
        self.project_config['ppas'] = ppas
        logging.debug('PPAs: {0}'.format(', '.join(ppas)))


        # Page won't be completed until valudation is passed on every row
        invalid_ppas = any([color == 'red' for ppa_name, color in self.ppas])
        self.assistant.set_page_complete(self.ppas_vbox,
                                         len(ppas) > 0 and not invalid_ppas)


    def version_checkbutton_toggled_cb(self, checkbutton):
        self._update_target_versions()


    def _update_target_versions(self):
        """
        Update target versions based on checkbuttons status
        """
        version_template = self._update_target_version_helper()

        # Branch will be tagged if revision number isn't appended
        # to changelog version
        self.project_config['tag'] = \
            not self.append_revno_checkbutton.get_active()

        # Populate model with target version for each distro
        self.target_versions.clear()

        context = {'revno': self.bzr_info.revno(),
                   'date': datetime.now()}
        distros=self.project_config['distros']
        for distro in distros:
            context['distro'] = distro
            target_version = Template(version_template).render(**context)
            self.target_versions.append((target_version,))

        # Page will be complete when there isn't any duplicated target version
        target_versions = [model_data[0] for model_data in self.target_versions]
        duplicated = len(set(target_versions)) != len(target_versions)
        self.assistant.set_page_complete(self.versions_vbox, not duplicated)

        logging.debug('Tag: {0}'.format(self.project_config['tag']))
        logging.debug('Target versions: {0}'
                      .format(', '.join(target_versions)))


    def distro_toggled_cb(self, renderer, path):
        """
        Select/Deselect model in distro
        """
        self.distros[path][self.SELECTED] = not self.distros[path][self.SELECTED]

        # Page will be complete when at least one target distro has been selected
        selected_column = [selected for distro, selected in self.distros]
        self.assistant.set_page_complete(self.distros_vbox, any(selected_column))

        logging.debug('Target distributions: {0}'
                      .format(' '.join([distro
                                        for distro, selected in self.distros
                                        if selected])))
