import pygtk
pygtk.require("2.0")
import gtk

import sys, os, logging, threading, subprocess
from traceback import format_exception, extract_tb
from contextlib import contextmanager

from ... import interface


class ExceptionDialog(gtk.Dialog):
    """
    Dialog that shows information regarding an exception
    """
    CONTINUE = 1
    REPORT_BUG = 2

    def __init__(self, message=None, buttons=None, exc_data=None):
        # Log exception data before creating the dialog
        if exc_data is None:
            exc_data = sys.exc_info()
        exception_str = ''.join(format_exception(*exc_data))
        logging.error(exception_str)

        gtk.Dialog.__init__(self, 'Error',
                            flags = gtk.DIALOG_MODAL,
                            buttons = buttons)
        self.set_property('border-width', 10)
        self.vbox.set_spacing(5)

        if not buttons:
            self.add_button('Continue', self.CONTINUE)
            button = gtk.Button('Report problem...')
            image = gtk.image_new_from_icon_name('apport',
                                                 gtk.ICON_SIZE_BUTTON)
            button.set_image(image)
            self.add_action_widget(button, self.REPORT_BUG)

        self.set_default_response(self.CONTINUE)

        hbox = gtk.HBox(spacing=5)
        image = gtk.image_new_from_stock(gtk.STOCK_DIALOG_ERROR,
                                         gtk.ICON_SIZE_DIALOG)
        image.set_alignment(0.5, 0.0)
        label = gtk.Label()

        _type, value, traceback = exc_data
        if message:
            label.set_markup(message)
        elif isinstance(value, interface.PackagerException):
            label.set_markup(value.message)
        else:
            (filename, line_number, function_name, text) = extract_tb(traceback)[-1]
            label.set_markup('Unexpected exception in:\n'
                             '- File: <b>{0}</b>, line <b>{1}</b>\n'
                             '- Function: <b>{2}</b>'
                             .format(os.path.basename(filename),
                                     line_number,
                                     function_name))
        hbox.pack_start(image, expand=False)
        hbox.pack_start(label, expand=False)
        self.vbox.pack_start(hbox, expand=False)

        expander = gtk.expander_new_with_mnemonic('_More information')

        scrolled_window = gtk.ScrolledWindow()
        textview = gtk.TextView()
        textview.set_editable(False)
        buffer = textview.get_buffer()
        buffer.set_text(exception_str)
        scrolled_window.add(textview)

        expander.add(scrolled_window)
        self.vbox.pack_start(expander)


class ProgressDialog(gtk.Dialog):
    """
    Display a progress dialog while some long operation is performed
    """
    JOIN_TIME = 0.2

    def __init__(self, title, message, thread):
        super(ProgressDialog, self).__init__(title,
                                             flags=gtk.DIALOG_MODAL)
        self.set_property('border-width', 10)
        self.set_has_separator(False)
        self.vbox.set_spacing(5)

        hbox = gtk.HBox(spacing=5)
        image = gtk.image_new_from_stock(gtk.STOCK_DIALOG_INFO,
                                         gtk.ICON_SIZE_DIALOG)
        image.set_alignment(0.5, 0.0)
        hbox.pack_start(image)
        label = gtk.Label()
        label.set_markup(message)
        hbox.pack_start(label)
        self.vbox.pack_start(hbox)

        progress_bar = gtk.ProgressBar()
        self.vbox.pack_start(progress_bar)

        # Prevent dialog from being closed and resized
        self.connect('delete-event', lambda w, e: True)
        self.set_deletable(False)
        self.set_resizable(False)
        self.show_all()

        self.message=message
        self.thread=thread
        self.progress_bar=progress_bar


    def run(self):
        """
        Pulse progress bar while thread is runnin
        """
        logging.debug(self.message)
        def refresh():
            """
            Process pending events
            """
            while gtk.events_pending():
                gtk.main_iteration()

        refresh()

        self.thread.start()
        while True:
            self.thread.join(self.JOIN_TIME)
            if not self.thread.is_alive():
                break

            self.progress_bar.pulse()
            refresh()

        self.destroy()
        return self


class MessageDialog(gtk.MessageDialog):
    """
    Create a MessageDialog, run it, destroy it and return response
    """
    # Logging function used depending on dialog type
    TYPE_TO_LOGGING = {
        gtk.MESSAGE_INFO: logging.info,
        gtk.MESSAGE_WARNING: logging.warning,
        gtk.MESSAGE_QUESTION: logging.debug,
        gtk.MESSAGE_ERROR: logging.error,
        gtk.MESSAGE_OTHER: logging.debug,
        }

    def __init__(self, title, message,
                 type=gtk.MESSAGE_ERROR,
                 buttons=gtk.BUTTONS_OK,
                 flags=gtk.DIALOG_MODAL,
                 default_response=None,
                 response_cb=None):
        """
        Create MessageDialog object
        """
        (super(MessageDialog, self)
         .__init__(type=type, buttons=buttons, flags=flags))
        self.set_title(title)
        self.set_markup(message)
        if default_response:
            self.set_default_response(default_response)

        def default_response_cb(dialog, response):
            dialog.destroy()

        self.connect('response', response_cb or default_response_cb)

        self.message = message
        self.message_type = type


    def run(self):
        """
        Run dialog, destroy it and return response
        """
        self.TYPE_TO_LOGGING[self.message_type](self.message)
        self.show_all()


class Thread(threading.Thread):
    """
    Thread that keeps what target function returns
    in target_return variable
    """
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs=None, verbose=None):
        if kwargs is None:
            kwargs = {}
        super(Thread, self).__init__(group, target, name,
                                     args, kwargs, verbose)
        def target_wrapper():
            self.target_return = target(*args, **kwargs)

        # Keep original values for logging purposes
        self.target = target
        self.args = args
        self.kwargs = kwargs

        # Replace values with wrapper data
        self.__target = target_wrapper
        self.__args = ()
        self.__kwargs = {}


    def start(self):
        """
        Execute thread
        """
        logging.debug('Starting thread: {0}, {1}, {2}'
                      .format(self.target, self.args, self.kwargs))
        super(Thread, self).start()


class Command(object):
    """
    Execute command in another process and log its output
    """
    def __init__(self, command_data):
        if isinstance(command_data, str):
            shell=True
            self.command_str = command_data
        elif (isinstance(command_data, list)
              or isinstance(command_data, tuple)):
            shell=False
            self.command_str = ' '.join(command_data)
        else:
            raise ValueError(command_data)

        self.stdout = None
        self.stderr = None
        self.process = subprocess.Popen(command_data,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        shell=shell)

    def summary(self):
        """
        Return a summary of the command execution
        """
        text = [self.command_str,
                ('-returncode: {0}'
                 .format(self.process.returncode))]
        if self.stdout:
            text.append('-stdout:\n{0}'.format(self.stdout))
        if self.stderr:
            text.append('-stderr:\n{0}'.format(self.stderr))

        return '\n'.join(text)


    def run(self):
        """
        Run command and log its output
        """
        logging.debug('Running {0!r}...'.format(self.command_str))
        self.stdout, self.stderr = self.process.communicate()
        logging.debug(self.summary())
        return self


@contextmanager
def block_handler(widget, func):
    widget.handler_block_by_func(func)
    yield
    widget.handler_unblock_by_func(func)
